@@ -17,19 +17,21 @@
  * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
  */
 
-#ifndef ERROR_H
-#define ERROR_H
+#ifndef COMMON_ERROR_H
+#define COMMON_ERROR_H
 
 #include <stdarg.h> 	/* for ... */
 #include <errno.h>  	/* for errstr */
 #include <string.h> 	/* for strerror */
+#include <stdio.h>	/* for FILE */
 
 /* Current error message as a string. */
-#define errstr strerror(errno)
+#define errstr (strerror(errno))
 
-typedef int (*printf_fn_t)(const char *format, va_list ap);
+typedef int (*vprintf_fn_t)(const char *format, va_list ap);
+typedef int (*vfprintf_fn_t)(FILE *stream, const char *format, va_list ap);
 
-extern printf_fn_t warn_writer;
+extern vprintf_fn_t warn_writer;
 
 int default_warn_writer(const char *format, va_list ap);
 void die(const char *format, ...) __attribute__ ((noreturn, format (printf, 1, 2)));
