@@ -32,15 +32,16 @@
 #include "minmax.h"		/* Gnulib */
 #include "full-write.h"		/* Gnulib */
 #include "xvasprintf.h"		/* Gnulib */
+#include "quotearg.h"		/* Gnulib */
 #include "gettext.h"		/* Gnulib/GNU gettext */
 #define _(s) gettext(s)
 #define N_(s) gettext_noop(s)
-#include "common/common.h"
+#include "common/minmaxonce.h"
 #include "common/error.h"
-#include "common/substrcmp.h"
 #include "common/intutil.h"
 #include "common/strbuf.h"
 #include "common/strleftcmp.h"
+#include "common/comparison.h"
 #include "microdc.h"
 
 #define MAX_DIR_DEPTH 32
@@ -76,7 +77,7 @@ filelist_free(DCFileList *node)
     case DC_TYPE_REG:
     	break;
     case DC_TYPE_DIR:
-    	hmap_foreach(node->u.dir.children, filelist_free);
+    	hmap_foreach_value(node->u.dir.children, filelist_free);
 	hmap_free(node->u.dir.children);
     	break;
     }
@@ -160,8 +161,8 @@ parse_decoded_dclst(char *decoded, uint32_t decoded_len)
 DCFileList *
 filelist_lookup(DCFileList *node, const char *filename)
 {
-    HMapIterator it;
     const char *end;
+    char *name;
 
     if (*filename != '/')
     	return NULL;
@@ -174,19 +175,19 @@ filelist_lookup(DCFileList *node, const char *filename)
     }
     if (node->type != DC_TYPE_DIR)
     	return NULL;
-    if (substrcmp("..", filename, end-filename) == 0) {
+    if (strncmp(filename, "..", end-filename) == 0) {
     	DCFileList *parent = (node->parent == NULL ? node : node->parent);
        	return (*end == '\0' ? parent : filelist_lookup(parent, end));
     }
-    if (substrcmp(".", filename, end-filename) == 0)
+    if (strncmp(filename, ".", end-filename) == 0)
     	return (*end == '\0' ? node : filelist_lookup(node, end));
 
-    hmap_iterator(node->u.dir.children, &it);
-    while (it.has_next(&it)) {
-    	DCFileList *subnode = it.next(&it);
-    	if (substrcmp(subnode->name, filename, end-filename) == 0)
-	    return (*end == '\0' ? subnode : filelist_lookup(subnode, end));
-    }
+    /* FIXME. what's this? Why not do a real lookup instead?! */
+    name = strndup(filename, end-filename);
+    node = hmap_get(node->u.dir.children, name);
+    free(name);
+    if (node != NULL)
+	return (*end == '\0' ? node : filelist_lookup(node, end));
 
     return NULL;
 }
@@ -278,7 +279,7 @@ filelist_list_recursively(DCFileList *node, char *basepath, int skip)
 	char *tmp;
 
 	tmp = catfiles(basepath, node->name);
-	screen_putf("%7" PRIu64 "M %s%s\n", node->u.reg.size/(1024*1024), tmp, "");
+	screen_putf("%7" PRIu64 "M %s%s\n", node->u.reg.size/(1024*1024), quotearg(tmp), "");
 	free(tmp);
     }
 }
@@ -326,16 +327,16 @@ filelist_list(DCFileList *node, bool long_mode)
 	    for (c = 0; items[c] != NULL; c++) {
 		switch (items[c]->type) {
 		case DC_TYPE_REG:
-    		    screen_putf(format, items[c]->u.reg.size/(1024*1024), items[c]->name, "");
+    		    screen_putf(format, items[c]->u.reg.size/(1024*1024), quotearg(items[c]->name), "");
 		    break;
 		case DC_TYPE_DIR:
-    		    screen_putf(format, items[c]->u.dir.totalsize/(1024*1024), items[c]->name, "/");
+    		    screen_putf(format, items[c]->u.dir.totalsize/(1024*1024), quotearg(items[c]->name), "/");
 		    break;
 		}
 	    }
 	    free(items);
 	} else {
-	    screen_putf(format, node->u.reg.size, node->name, "");
+	    screen_putf(format, node->u.reg.size, quotearg(node->name), "");
 	}
 	free(format);
     } else {
@@ -363,11 +364,11 @@ filelist_list(DCFileList *node, bool long_mode)
 		    switch (item->type) {
 		    case DC_TYPE_REG:
 			extlen = 0;
-			screen_putf("%s", item->name);
+			screen_putf("%s", quotearg(item->name));
 			break;
 		    case DC_TYPE_DIR:
 			extlen = 1;
-			screen_putf("%s/", item->name);
+			screen_putf("%s/", quotearg(item->name));
 			break;
 		    }
 		    if (c+rows < count) {
@@ -379,7 +380,7 @@ filelist_list(DCFileList *node, bool long_mode)
 	    }
 	    free(items);
 	} else {
-    	    screen_putf("%s\n", node->name);
+    	    screen_putf("%s\n", quotearg(node->name));
 	}
     }
 }
@@ -396,36 +397,36 @@ filelist_open(const char *filename)
     ssize_t res;
 
     if (stat(filename, &st) < 0) {
-    	screen_putf(_("%s: Cannot get file status - %s\n"), filename, errstr);
+    	screen_putf(_("%s: Cannot get file status - %s\n"), quotearg(filename), errstr);
 	return NULL;
     }
     contents = malloc(st.st_size);
     if (contents == NULL) {
-    	screen_putf("%s: %s\n", filename, errstr);
+    	screen_putf("%s: %s\n", quotearg(filename), errstr);
 	return NULL;
     }
     fd = open(filename, O_RDONLY);
     if (fd < 0) {
-    	screen_putf(_("%s: Cannot open file for reading - %s\n"), filename, errstr);
+    	screen_putf(_("%s: Cannot open file for reading - %s\n"), quotearg(filename), errstr);
 	return NULL;
     }
     res = full_read(fd, contents, st.st_size);
     if (res < st.st_size) {
     	if (res < 0)
-    	    screen_putf(_("%s: Cannot read from file - %s\n"), filename, errstr);
+    	    screen_putf(_("%s: Cannot read from file - %s\n"), quotearg(filename), errstr);
 	else
-	    screen_putf(_("%s: Premature end of file\n"), filename);	/* XXX: really: file was truncated? */
+	    screen_putf(_("%s: Premature end of file\n"), quotearg(filename));	/* XXX: really: file was truncated? */
 	free(contents);
 	if (close(fd) < 0)
-    	    screen_putf(_("%s: Cannot close file - %s\n"), filename, errstr);
+    	    screen_putf(_("%s: Cannot close file - %s\n"), quotearg(filename), errstr);
 	return NULL;
     }
     if (close(fd) < 0)
-    	screen_putf(_("%s: Cannot close file - %s\n"), filename, errstr);
+    	screen_putf(_("%s: Cannot close file - %s\n"), quotearg(filename), errstr);
     decoded = huffman_decode(contents, st.st_size, &decoded_len);
     free(contents);
     if (decoded == NULL) {
-    	screen_putf(_("%s: Invalid data, cannot decode\n"), filename);
+    	screen_putf(_("%s: Invalid data, cannot decode\n"), quotearg(filename));
 	return NULL;
     }
     root = parse_decoded_dclst(decoded, decoded_len);
@@ -441,7 +442,7 @@ dir_to_filelist(DCFileList *parent, const char *path)
 
     dp = opendir(path);
     if (dp == NULL) {
-    	screen_putf(_("%s: Cannot open directory - %s\n"), path, errstr);
+    	screen_putf(_("%s: Cannot open directory - %s\n"), quotearg(path), errstr);
 	return;
     }
     errno = 0;
@@ -456,7 +457,7 @@ dir_to_filelist(DCFileList *parent, const char *path)
 
     	fullname = catfiles(path, ep->d_name);
     	if (stat(fullname, &st) < 0) {
-	    screen_putf(_("%s: Cannot get file status - %s\n"), fullname, errstr);
+	    screen_putf(_("%s: Cannot get file status - %s\n"), quotearg(fullname), errstr);
 	    free(fullname);
 	    continue;
 	}
@@ -475,14 +476,14 @@ dir_to_filelist(DCFileList *parent, const char *path)
 	    hmap_put(parent->u.dir.children, node->name, node);
 	}
 	else {
-	    screen_putf(_("%s: Not a regular file or directory, ignoring\n"), fullname);
+	    screen_putf(_("%s: Not a regular file or directory, ignoring\n"), quotearg(fullname));
 	}
 	free(fullname);
     }
     if (errno != 0)
-    	screen_putf(_("%s: Cannot read directory - %s\n"), path, errstr);
+    	screen_putf(_("%s: Cannot read directory - %s\n"), quotearg(path), errstr);
     if (closedir(dp) < 0)
-    	screen_putf(_("%s: Cannot close directory - %s\n"), path, errstr);
+    	screen_putf(_("%s: Cannot close directory - %s\n"), quotearg(path), errstr);
 }
 
 static void
@@ -523,7 +524,7 @@ filelist_create(const char *basedir)
 
     sb = strbuf_new();
     if (basedir != NULL) {
-    	screen_putf(_("Scanning directory %s\n"), basedir);
+    	screen_putf(_("Scanning directory %s\n"), quotearg(basedir));
     	dir_to_filelist(root, basedir);
 	filelist_to_string(root, sb, 0);
     }
@@ -548,7 +549,7 @@ filelist_create(const char *basedir)
 
     fd = open(filename, O_CREAT|O_EXCL|O_WRONLY, 0666);
     if (fd < 0) {
-    	screen_putf(_("%s: Cannot open file for writing - %s\n"), filename, errstr);
+    	screen_putf(_("%s: Cannot open file for writing - %s\n"), quotearg(filename), errstr);
 	filelist_free(root);
 	free(outdata);
 	free(filename);
@@ -558,14 +559,14 @@ filelist_create(const char *basedir)
     	ptrv_append(delete_files, xstrdup(filename));
 
     if (full_write(fd, outdata, len) < len) {
-    	screen_putf(_("%s: Cannot write to file - %s\n"), filename, errstr);
+    	screen_putf(_("%s: Cannot write to file - %s\n"), quotearg(filename), errstr);
 	filelist_free(root); 
 	free(outdata);
 	free(filename);
 	return false;
     }
     if (close(fd) < 0)
-    	screen_putf(_("%s: Cannot close file - %s\n"), filename, errstr);
+    	screen_putf(_("%s: Cannot close file - %s\n"), quotearg(filename), errstr);
     free(filename);
     free(outdata);
 
@@ -737,9 +738,8 @@ local_fs_completion_generator(const char *dir_part, const char *file_part, int s
 		    	/*ce->display_char = '\0';*/
 		    }
 		}
-		ce->str = xstrdup(de->d_name);
+		ce->str = xquotestr(de->d_name);
 		free(full);
-
 		return true;
 	    }
 	}
@@ -812,7 +812,7 @@ remote_fs_completion_generator(const char *dir_path, const char *file_part, int
 	    continue;
 
 	if (strleftcmp(file_part, node->name) == 0) {
-    	    ce->str = xstrdup(node->name);
+    	    ce->str = xquotestr(node->name);
 	    if (node->type == DC_TYPE_DIR) {
 		ce->input_char = '/';
 		/*ce->display_char = '/';*/
