@@ -28,6 +28,7 @@
 #include "xvasprintf.h"		/* Gnulib */
 #include "xstrndup.h"		/* Gnulib */
 #include "xalloc.h"		/* Gnulib */
+#include "quotearg.h"		/* Gnulib */
 #include "gettext.h"		/* Gnulib/GNU gettext */
 #define _(s) gettext(s)
 #define N_(s) gettext_noop(s)
@@ -63,7 +64,7 @@ say_user_completion_generator(const char *base, int state, DCCompletionEntry *ce
     while (it.has_next(&it)) {
     	DCUserInfo *ui = it.next(&it);
 	if (strleftcmp(base, ui->nick) == 0) {
-	    ce->str = xstrdup(ui->nick);
+	    ce->str = xquotestr(ui->nick);
 	    ce->input_single_format = "%s:";
 	    /*ce->input_append_single = ':';*/
 	    return true;
@@ -81,16 +82,15 @@ user_or_myself_completion_generator(const char *base, int state)
     if (state == 0) {
     	hmap_iterator(hub_users, &it);
 	if (strleftcmp(base, my_nick) == 0)
-	    return xstrdup(my_nick);
+	    return xquotestr(my_nick);
     }
 
     while (it.has_next(&it)) {
     	DCUserInfo *ui = it.next(&it);
 	if (strleftcmp(base, ui->nick) == 0)
-	    return xstrdup(ui->nick);
+	    return xquotestr(ui->nick);
     }
 
-
     return NULL;
 }
 
@@ -105,7 +105,24 @@ user_completion_generator(const char *base, int state)
     while (it.has_next(&it)) {
     	DCUserInfo *ui = it.next(&it);
 	if (strleftcmp(base, ui->nick) == 0)
-	    return xstrdup(ui->nick);
+	    return xquotestr(ui->nick);
+    }
+
+    return NULL;
+}
+
+char *
+user_with_queue_completion_generator(const char *base, int state)
+{
+    static HMapIterator it;
+
+    if (state == 0)
+    	hmap_iterator(hub_users, &it);
+
+    while (it.has_next(&it)) {
+    	DCUserInfo *ui = it.next(&it);
+	if (ui->download_queue->cur > 0 && strleftcmp(base, ui->nick) == 0)
+	    return xquotestr(ui->nick);
     }
 
     return NULL;
@@ -249,7 +266,7 @@ hub_disconnect(void)
     	hub_socket = -1;
     }
     if (hub_users != NULL) {
-	hmap_foreach(hub_users, user_info_free);
+	hmap_foreach_value(hub_users, user_info_free);
 	hmap_clear(hub_users);
     }
     if (hub_sendq != NULL)
@@ -257,7 +274,7 @@ hub_disconnect(void)
     if (hub_recvq != NULL)
     	byteq_clear(hub_recvq);
     if (pending_userinfo != NULL) {
-    	hmap_foreach(pending_userinfo, user_info_free);
+    	hmap_foreach_value(pending_userinfo, user_info_free);
     	hmap_clear(pending_userinfo);
     }
     hub_state = DC_HUB_DISCONNECTED;
@@ -285,7 +302,7 @@ hub_handle_command(char *buf, uint32_t len)
 
         key = memmem(buf+6, len-6, " Pk=", 4);
         if (key == NULL) {
-            screen_putf(_("Invalid $Lock message: Missing Pk value\n"));
+            warn(_("Invalid $Lock message: Missing Pk value\n"));
 	    key = buf+len;
 	}
         key = decode_lock(buf+6, key-buf-6, DC_CLIENT_BASE_KEY);
@@ -309,7 +326,7 @@ hub_handle_command(char *buf, uint32_t len)
 	    return;
     }
     else if (strcmp(buf, "$BadPass") == 0) {
-    	screen_putf(_("Password not accepted.\n"));
+    	warn(_("Password not accepted.\n"));
 	hub_disconnect();
     }
     else if (strcmp(buf, "$LogedIn") == 0) {
@@ -318,7 +335,10 @@ hub_handle_command(char *buf, uint32_t len)
     else if (len >= 9 && strncmp(buf, "$HubName ", 9) == 0) {
 	free(hub_name);
 	hub_name = xstrdup(buf+9);
-        screen_putf(_("Hub name is %s.\n"), hub_name);
+        screen_putf(_("Hub name is %s.\n"), quotearg(hub_name));
+    }
+    else if (strcmp(buf, "$GetNetInfo") == 0) {
+	hub_putf("$NetInfo %d$1$%c|", my_ul_slots, is_active ? 'A' : 'P');
     }
     else if (strcmp(buf, "$ValidateDenide") == 0) {
     	if (!check_state(buf, DC_HUB_HELLO))
@@ -327,7 +347,7 @@ hub_handle_command(char *buf, uint32_t len)
          * But shouldn't we give the client a chance to change the nick?
          * Also what happens if we receive this when correctly logged in?
          */
-    	screen_putf(_("Hub did not accept nick. Nick may be in use.\n"));
+    	warn(_("Hub did not accept nick. Nick may be in use.\n"));
     	hub_disconnect();
     }
     else if (len >= 7 && strncmp(buf, "$Hello ", 7) == 0) {
@@ -340,7 +360,7 @@ hub_handle_command(char *buf, uint32_t len)
 	    	/* This probably won't happen, but better safe... */
 	    	free(my_nick);
 		my_nick = xstrdup(buf+7);
-    	    	screen_putf(_("Nick accepted but modified to %s. You are now logged in.\n"), my_nick);
+    	    	screen_putf(_("Nick accepted but modified to %s. You are now logged in.\n"), quotearg(my_nick));
 	    }
 
     	    ui = user_info_new(my_nick);
@@ -356,7 +376,7 @@ hub_handle_command(char *buf, uint32_t len)
 	    	return;
     	    hub_state = DC_HUB_LOGGED_IN;
 	} else {
-	    flag_putf(DC_DF_JOIN_PART, _("User %s logged in.\n"), buf+7);
+	    flag_putf(DC_DF_JOIN_PART, _("User %s logged in.\n"), quotearg(buf+7));
 	    ui = user_info_new(buf+7);
             hmap_put(hub_users, ui->nick, ui);
 	    if (!hub_putf("$GetINFO %s|", buf+7))
@@ -371,24 +391,24 @@ hub_handle_command(char *buf, uint32_t len)
     	buf += 8;
     	token = strsep(&buf, " ");
 	if (strcmp(token, "$ALL") != 0) {
-	    screen_putf(_("Invalid $MyINFO message: Missing $ALL parameter, ignoring\n"));
+	    warn(_("Invalid $MyINFO message: Missing $ALL parameter, ignoring\n"));
 	    return;
 	}
 
     	token = strsep(&buf, " ");
 	if (token == NULL) {
-	    screen_putf(_("Invalid $MyINFO message: Missing nick parameter, ignoring\n"));
+	    warn(_("Invalid $MyINFO message: Missing nick parameter, ignoring\n"));
 	    return;
 	}
     	ui = hmap_get(hub_users, token);
 	if (ui == NULL) {
-	    screen_putf(_("Invalid $MyINFO message: Unknown user %s, ignoring\n"), token);
+	    warn(_("Invalid $MyINFO message: Unknown user %s, ignoring\n"), quotearg(token));
 	    return;
 	}
 
     	token = strsep(&buf, "$");
 	if (token == NULL) {
-	    screen_putf(_("Invalid $MyINFO message: Missing description parameter, ignoring\n"));
+	    warn(_("Invalid $MyINFO message: Missing description parameter, ignoring\n"));
 	    return;
 	}
 	free(ui->description);
@@ -396,13 +416,13 @@ hub_handle_command(char *buf, uint32_t len)
 
     	token = strsep(&buf, "$");
 	if (token == NULL) {
-	    screen_putf(_("Invalid $MyINFO message: Missing description separator, ignoring\n"));
+	    warn(_("Invalid $MyINFO message: Missing description separator, ignoring\n"));
 	    return;
 	}
 
     	token = strsep(&buf, "$");
 	if (token == NULL) {
-	    screen_putf(_("Invalid $MyINFO message: Missing connection speed, ignoring\n"));
+	    warn(_("Invalid $MyINFO message: Missing connection speed, ignoring\n"));
 	    return;
 	}
 	len = strlen(token);
@@ -417,7 +437,7 @@ hub_handle_command(char *buf, uint32_t len)
 
     	token = strsep(&buf, "$");
 	if (token == NULL) {
-	    screen_putf(_("Invalid $MyINFO message: Missing e-mail address, ignoring\n"));
+	    warn(_("Invalid $MyINFO message: Missing e-mail address, ignoring\n"));
 	    return;
 	}
 	free(ui->email);
@@ -425,11 +445,11 @@ hub_handle_command(char *buf, uint32_t len)
 
     	token = strsep(&buf, "$");
 	if (token == NULL) {
-	    screen_putf(_("Invalid $MyINFO message: Missing share size, ignoring\n"));
+	    warn(_("Invalid $MyINFO message: Missing share size, ignoring\n"));
 	    return;
 	}
 	if (!parse_uint64(token, &ui->share_size)) {
-	    screen_putf(_("Invalid $MyINFO message: Invalid share size, ignoring\n"));
+	    warn(_("Invalid $MyINFO message: Invalid share size, ignoring\n"));
 	    return;
 	}
 
@@ -439,39 +459,84 @@ hub_handle_command(char *buf, uint32_t len)
 	/* XXX: Now that user's active_state may have changed, try queue again? */
     }
     else if (strcmp(buf, "$HubIsFull") == 0) {
-        screen_putf(_("Hub is full.\n"));
+        warn(_("Hub is full.\n"));
         /* DC++ does not disconnect immediately. So I guess we won't either. */
 	/* Maybe we should be expecting an "hub follow" message? */
 	/* hub_disconnect(); */
     }
     else if (len >= 3 && (buf[0] == '<' || strncmp(buf, " * ", 3) == 0)) {
-        flag_putf(DC_DF_PUBLIC_CHAT, _("Public: %s\n"), buf);
+	char *end;
+	bool first = true;
+	/*int scrwidth;
+	size_t firstlen;
+	size_t otherlen;
+
+        screen_get_size(NULL, &scrwidth);
+        firstlen = scrwidth - strlen(_("Public:"));
+        otherlen = scrwidth - strlen(_(" | "));*/
+	for (; (end = strchr(buf, '\n')) != NULL; buf = end+1) {
+	    /*PtrV *wrapped;*/
+
+	    if (end[-1] == '\r') /* end > buf here, buf[0] == '<' or ' ' */
+		end[-1] = '\0';
+	    else
+		end[0] = '\0';
+
+            /*wrapped = wordwrap(quotearg(buf), first ? firstlen : otherlen, otherlen);
+            for (c = 0; c < wrapped->cur; c++)
+	        flag_putf(DC_DF_PUBLIC_CHAT, first ? _("Public: %s\n") : _(" | %s\n"), );
+            ptrv_foreach(wrapped, free);
+            ptrv_free(wrapped);*/
+	    flag_putf(DC_DF_PUBLIC_CHAT, first ? _("Public: %s\n") : _(" | %s\n"), quotearg(buf));
+            first = false;
+	}
+	flag_putf(DC_DF_PUBLIC_CHAT, first ? _("Public: %s\n") : _(" | %s\n"), quotearg(buf));
     }
     else if (len >= 5 && strncmp(buf, "$To: ", 5) == 0) {
     	char *msg;
+	char *end;
+	bool first = true;
+
 	msg = strchr(buf+5, '$');
 	if (msg == NULL) {
-	    screen_putf(_("Invalid $To message: Missing text separator, ignoring\n"));
+	    warn(_("Invalid $To message: Missing text separator, ignoring\n"));
 	    return;
         }
         *msg = '\0';
 	msg++;
-        screen_putf(_("Private: [%s] %s\n"), buf+5, msg);
+
+	for (; (end = strchr(msg, '\n')) != NULL; msg = end+1) {
+	    if (end[-1] == '\r') /* end > buf here, msg[0] == '<' or ' ' */
+		end[-1] = '\0';
+	    else
+		end[0] = '\0';
+	    if (first) {
+		screen_putf(_("Private: [%s] %s\n"), quotearg_n(0, buf+5), quotearg_n(1, msg));
+		first = false;
+	    } else {
+		screen_putf(_(" | %s\n"), quotearg(msg));
+	    }
+	}
+	if (first) {
+	    screen_putf(_("Private: [%s] %s\n"), quotearg_n(0, buf+5), quotearg_n(1, msg));
+	} else {
+	    screen_putf(_(" | %s\n"), quotearg(msg));
+	}
     }
     else if (len >= 13 && strncmp(buf, "$ConnectToMe ", 13) == 0) {
 	struct sockaddr_in addr;
 
     	buf += 13;
 	if (strsep(&buf, " ") == NULL) {
-	    screen_putf(_("Invalid $ConnectToMe message: Missing or invalid nick\n"));
+	    warn(_("Invalid $ConnectToMe message: Missing or invalid nick\n"));
 	    return;
     	}
     	if (!parse_ip_and_port(buf, &addr, 0)) {
-	    screen_putf(_("Invalid $ConnectToMe message: Invalid address specification.\n"));
+	    warn(_("Invalid $ConnectToMe message: Invalid address specification.\n"));
 	    return;
         }
 
-	screen_putf(_("Connecting to user on %s\n"), sockaddr_in_str(&addr));
+	flag_putf(DC_DF_CONNECTIONS, _("Connecting to user on %s\n"), sockaddr_in_str(&addr));
     	user_connection_new(&addr, -1);
     }
     else if (len >= 16 && strncmp(buf, "$RevConnectToMe ", 16) == 0) {
@@ -480,27 +545,27 @@ hub_handle_command(char *buf, uint32_t len)
 
     	nick = strtok(buf+16, " ");
 	if (nick == NULL) {
-	    screen_putf(_("Invalid $RevConnectToMe message: Missing nick parameter\n"));
+	    warn(_("Invalid $RevConnectToMe message: Missing nick parameter\n"));
 	    return;
         }
     	if (strcmp(nick, my_nick) == 0) {
-	    screen_putf(_("Invalid $RevConnectToMe message: Remote nick is our nick\n"));
+	    warn(_("Invalid $RevConnectToMe message: Remote nick is our nick\n"));
 	    return;
 	}
         ui = hmap_get(hub_users, nick);
         if (ui == NULL) {
-	    screen_putf(_("Invalid $RevConnectToMe message: Unknown user %s, ignoring\n"), nick);
+	    warn(_("Invalid $RevConnectToMe message: Unknown user %s, ignoring\n"), quotearg(nick));
 	    return;
         }
 
 	if (ui->conn_count >= DC_USER_MAX_CONN) {
-	    screen_putf(_("No more connections to user %s allowed.\n"), ui->nick);
+	    warn(_("No more connections to user %s allowed.\n"), quotearg(ui->nick));
 	    return;
 	}
 
     	if (!is_active) {
     	    if (ui->active_state == DC_ACTIVE_SENT_PASSIVE) {
-    		screen_putf(_("User %s is also passive. Cannot establish connection.\n"), ui->nick);
+    		warn(_("User %s is also passive. Cannot establish connection.\n"), quotearg(ui->nick));
 		/* We could set this to DC_ACTIVE_UNKNOWN too. This would mean
 		 * we would keep sending RevConnectToMe next time the download
 		 * queue is modified or some timer expired and told us to retry
@@ -550,10 +615,10 @@ hub_handle_command(char *buf, uint32_t len)
     	char *nick;
 	char *end;
 
-	for (nick = buf+10; (end = strstr(nick, "$$")) != NULL; nick = end+2) {
+	for (nick = buf+8; (end = strstr(nick, "$$")) != NULL; nick = end+2) {
             DCUserInfo *ui;
 
-    	    *end = '\0';
+   	    *end = '\0';
 	    ui = hmap_get(hub_users, nick);
 	    if (ui == NULL) {
                 ui = user_info_new(nick);
@@ -569,13 +634,13 @@ hub_handle_command(char *buf, uint32_t len)
     }
     else if (len >= 6 && strncmp(buf, "$Quit ", 6) == 0) {
     	DCUserInfo *ui;
-    	flag_putf(DC_DF_JOIN_PART, "User %s quits.\n", buf+6);
+    	flag_putf(DC_DF_JOIN_PART, "User %s quits.\n", quotearg(buf+6));
 	ui = hmap_remove(hub_users, buf+6);
 	if (ui == NULL) {
 	    /* Some hubs print quit messages for users that never joined,
 	     * so print this as debug only.
 	     */
-	    flag_putf(DC_DF_DEBUG, _("Invalid $Quit message: Unknown user %s.\n"), buf+6);
+	    flag_putf(DC_DF_DEBUG, _("Invalid $Quit message: Unknown user %s.\n"), quotearg(buf+6));
 	    return;
 	}
 	user_info_free(ui);
@@ -587,17 +652,17 @@ hub_handle_command(char *buf, uint32_t len)
     	buf += 8;
 	source = strsep(&buf, " ");
 	if (source == NULL) {
-	    screen_putf(_("Invalid $Search message: Missing source specification.\n"));
+	    warn(_("Invalid $Search message: Missing source specification.\n"));
 	    return;
 	}
 	if (!parse_search_selection(buf, &sel)) {
-	    screen_putf(_("Invalid $Search message: Invalid search specification.\n"));
+	    warn(_("Invalid $Search message: Invalid search specification.\n"));
 	    return;
 	}
 	if (strncmp(source, "Hub:", 4) == 0) {
 	    DCUserInfo *ui = hmap_get(hub_users, source+4);
 	    if (ui == NULL) {
-		screen_putf(_("Invalid $Search message: Unknown user %s.\n"), source+4);
+		warn(_("Invalid $Search message: Unknown user %s.\n"), quotearg(source+4));
 		free(sel.patterns);
 		return;
 	    }
@@ -607,7 +672,7 @@ hub_handle_command(char *buf, uint32_t len)
 	} else {
             struct sockaddr_in addr;
     	    if (!parse_ip_and_port(source, &addr, DC_CLIENT_UDP_PORT)) {
-		screen_putf(_("Invalid $Search message: Invalid address specification.\n"));
+		warn(_("Invalid $Search message: Invalid address specification.\n"));
 		free(sel.patterns);
 		return;
 	    }
@@ -740,11 +805,11 @@ hub_connect_user(DCUserInfo *ui)
 	ui->active_state = DC_ACTIVE_SENT_ACTIVE;
     } else {
     	if (ui->active_state == DC_ACTIVE_SENT_PASSIVE) {
-    	    screen_putf(_("RevConnectToMe already sent to user %s. Waiting.\n"), ui->nick);
+    	    warn(_("RevConnectToMe already sent to user %s. Waiting.\n"), quotearg(ui->nick));
 	    return true;
 	}
     	if (ui->active_state == DC_ACTIVE_RECEIVED_PASSIVE) {
-    	    screen_putf(_("User %s is also passive. Cannot communicate.\n"), ui->nick);
+    	    warn(_("User %s is also passive. Cannot communicate.\n"), quotearg(ui->nick));
 	    return true;
 	}
     	if (!hub_putf("$RevConnectToMe %s %s|", my_nick, ui->nick))
