@@ -21,13 +21,14 @@
 #include <stdint.h>		/* Gnulib/POSIX/C99 */
 #include <stdlib.h>		/* C89 */
 #include <stdio.h>		/* C89 */
+#include "xalloc.h"		/* Gnulib */
 #include "gettext.h"		/* Gnulib/GNU gettext */
 #define _(s) gettext(s)
 #define N_(s) gettext_noop(s)
 #include "common/error.h"
 #include "common/strbuf.h"
 #include "common/ptrv.h"
-#include "xalloc.h"
+#include "common/comparison.h"
 #include "microdc.h"
 
 #define BIT_POS_INIT(pos)      ((pos)*8)
