@@ -26,9 +26,46 @@
 #include "full-write.h"		/* Gnulib */
 #include "full-read.h"		/* Gnulib */
 #include "xalloc.h"		/* Gnulib */
-#include "ipc_internal.h"
 #include "ipc.h"
 
+typedef struct _IPCMessageHeader IPCMessageHeader;
+typedef union _IPCArgValue IPCArgValue;
+typedef union _IPCArgRef IPCArgRef;
+
+union _IPCArgRef {
+    uint64_t *t_int64;
+    uint32_t *t_int32;
+    int *t_int;
+    bool *t_bool;
+    char **t_str;
+};
+
+union _IPCArgValue {
+    uint64_t t_int64;
+    uint32_t t_int32;
+    int t_int;
+    bool t_bool;
+    struct {
+        ssize_t len;
+        char *chars;
+    } t_str;
+};
+
+struct _IPCMessageHeader {
+    uint32_t id;
+    uint32_t size;
+    uint32_t arg_count;
+};
+
+struct _IPC {
+    int fd;
+    void *userdata;
+    bool has_header;
+    IPCMessageHeader header;
+    IPCMessageFunction handle_message;
+    IPCErrorFunction handle_error;
+};
+
 static bool
 ipc_read(IPC *ipc, void *data, uint32_t datalen)
 {
