@@ -36,11 +36,13 @@
 #include "xvasprintf.h"		/* Gnulib */
 #include "version-etc.h"	/* Gnulib */
 #include "xalloc.h"		/* Gnulib */
+#include "quotearg.h"		/* Gnulib */
 #include "gettext.h"		/* Gnulib/GNU gettext */
 #define _(s) gettext(s)
 #define N_(s) gettext_noop(s)
 #include "common/error.h"
 #include "common/strleftcmp.h"
+#include "common/tempdir.h"
 #include "ipc.h"
 #include "microdc.h"
 
@@ -88,11 +90,13 @@ char *my_password;
 bool is_active;
 uint64_t my_share_size = 0;
 uint32_t display_flags = ~(DC_DF_DEBUG); /* All flags except debug set */
+uint32_t log_flags = ~(DC_DF_DEBUG);
 
 /* interactive state */
 DCFileList *browse_list = NULL;
 DCUserInfo *browse_user = NULL;
 char *browse_path = NULL;
+char *browse_path_previous = NULL;
 bool browsing_myself = false;
 
 /* misc */
@@ -124,12 +128,12 @@ void
 warn_file_error(int res, bool write, const char *filename)
 {
     if (write) {
-	warn(_("%s: Cannot write to file - %s\n"), filename, errstr);
+	warn(_("%s: Cannot write to file - %s\n"), quotearg(filename), errstr);
     } else {
 	if (res < 0) {
-	    warn(_("%s: Cannot read from file - %s\n"), filename, errstr);
+	    warn(_("%s: Cannot read from file - %s\n"), quotearg(filename), errstr);
 	} else {
-	    warn(_("%s: Premature end of file\n"), filename);
+	    warn(_("%s: Premature end of file\n"), quotearg(filename));
 	}
     }
 }
@@ -138,12 +142,12 @@ void
 warn_socket_error(int res, bool write, const char *subject)
 {
     if (write) {
-	warn("Cannot send to %s - %s\n", subject, errstr);
+	warn("Cannot send to %s - %s\n", quotearg(subject), errstr);
     } else {
 	if (res < 0) {
-	    warn("Cannot receive from %s - %s\n", subject, errstr);
+	    warn("Cannot receive from %s - %s\n", quotearg(subject), errstr);
 	} else {
-	    warn("Disconnected from %s.\n", subject);
+	    warn("Disconnected from %s.\n", quotearg(subject));
 	}
     }
 }
@@ -170,7 +174,7 @@ validate_nick(DCUserConn *uc, const char *nick)
 	    user_info_free(ui);
 	    return true;
 	}
-        screen_putf(_("No more connections to user %s allowed.\n"), nick);
+        warn(_("No more connections to user %s allowed.\n"), quotearg(nick));
         hmap_remove(pending_userinfo, ui->nick);
 	user_info_free(ui);
 	return false;
@@ -184,7 +188,7 @@ validate_nick(DCUserConn *uc, const char *nick)
 	    uc->info->conn[uc->info->conn_count++] = uc;
 	    return true;
 	}
-        screen_putf(_("No more connections to user %s allowed.\n"), nick);
+        warn(_("No more connections to user %s allowed.\n"), quotearg(nick));
     }
 
     return false;
@@ -216,7 +220,7 @@ validate_direction(DCUserConn *uc, DCTransferDirection dir)
     ptrv_append(user_conn_unknown_free, uc->name);
     newname = xasprintf(
 	    dir == DC_DIR_SEND ? _("%s|UL") : _("%s|DL"), uc->info->nick);
-    screen_putf(_("User %s validated as %s\n"), uc->name, newname);
+    flag_putf(DC_DF_CONNECTIONS, _("User %s validated as %s\n"), quotearg_n(0, uc->name), quotearg_n(1, newname));
     uc->name = newname;
     uc->dir = dir;
     hmap_put(user_conns, uc->name, uc);
@@ -247,7 +251,7 @@ handle_user_ipc_error(IPC *ipc, IPCError error, int result, void *userdata)
 	break;
     case IPC_ERROR_DATA:
     	//XXX: assert not reached
-    	screen_putf(_("Logical communication error with user process.\n"));
+    	warn(_("Logical communication error with user process.\n"));
     	break;
     }
     hmap_remove(user_conns, uc->name);
@@ -264,7 +268,8 @@ handle_user_msg(IPC *ipc, uint32_t id, void *userdata)
     	char *msg;
 
     	ipc_get(ipc, id, IPC_STR, &msg, IPC_END);
-    	screen_putf(_("User %s: %s"), uc->name, msg);
+	/* Don't quote what the user prints */
+    	screen_putf(_("User %s: %s"), quotearg(uc->name), msg);
 	free(msg);
     	return true;
     }
@@ -370,14 +375,12 @@ handle_user_msg(IPC *ipc, uint32_t id, void *userdata)
 	    free(uc->transfer_file);
     	    uc->transfer_file = filename;
 	    uc->transfer_len = filesize;
-	    if (time(&uc->transfer_start) < 0) {
+	    if (time(&uc->transfer_start) == (time_t) -1) {
 	    	warn(_("Cannot get current time - %s\n"), errstr);
 		uc->transfer_start = 0;
 	    }
 	}
-	//screen_putf("main pre put DC_MSG_TRANSFER_STARTING\n");
 	reply = ipc_put(ipc, id, IPC_BOOL, reply, IPC_END);
-	//screen_putf("main post put DC_MSG_TRANSFER_STARTING\n");
 	return reply;
     }
     case DC_MSG_TRANSFER_FINISHED: {
@@ -398,10 +401,11 @@ handle_user_msg(IPC *ipc, uint32_t id, void *userdata)
 	    if (browse_user != NULL && strcmp(browse_user->nick, uc->info->nick) == 0 && browse_list == NULL) {
 		char *filename;
 
-		screen_putf(_("Now browsing %s.\n"), browse_user->nick);
+		screen_putf(_("Now browsing %s.\n"), quotearg(browse_user->nick));
 		filename = xasprintf("%s/%s.MyList.DcLst", listing_dir, browse_user->nick);
 		browse_list = filelist_open(filename);
 		browse_path = xstrdup("/");
+		browse_path_previous = NULL;
 		free(filename);
 		update_prompt();
 	    }
@@ -412,8 +416,9 @@ handle_user_msg(IPC *ipc, uint32_t id, void *userdata)
     	return ipc_get(uc->ipc, id, IPC_INT64, &uc->transfer_pos, IPC_END);
     }
     case DC_MSG_DISCONNECT:
+        /* The DISCONNECT message will not be sent from users, only to users. */
     default:
-        screen_putf(_("Received message %d from user process.\n"), id);
+        warn(_("Received unknown message %d from user process, shutting down process.\n"), id);
     	hmap_remove(user_conns, uc->name);
 	user_disconnect(uc);
 	return false;
@@ -437,7 +442,7 @@ add_user_conn(DCUserConn *uc)
 void
 user_conn_cancel(DCUserConn *uc)
 {
-    warn(_("Disconnecting user.\n"));
+    screen_putf(_("Disconnecting user.\n"));
 
     if (!ipc_put(uc->ipc, DC_MSG_DISCONNECT, IPC_END))
     	return;
@@ -470,7 +475,7 @@ transfer_completion_generator(const char *base, int state)
     while (it.has_next(&it)) {
     	DCUserConn *uc = it.next(&it);
 	if (strleftcmp(base, uc->name) == 0)
-	    return xstrdup(uc->name);
+	    return xquotestr(uc->name);
     }
 
     return NULL;
@@ -533,7 +538,7 @@ run_script(const char *filename, bool allow_missing)
     file = fopen(filename, "r");
     if (file == NULL) {
     	if (!allow_missing || errno != ENOENT)
-    	    screen_putf(_("%s: Cannot open file - %s\n"), filename, errstr);
+    	    warn(_("%s: Cannot open file - %s\n"), quotearg(filename), errstr);
     	return;
     }
     while (getline(&line, &line_len, file) >= 0) {
@@ -544,20 +549,22 @@ run_script(const char *filename, bool allow_missing)
     }
     free(line);
     if (ferror(file)) {
-    	screen_putf(_("%s: Cannot read from file - %s\n"), filename, errstr);
+    	warn(_("%s: Cannot read from file - %s\n"), quotearg(filename), errstr);
 	if (fclose(file) < 0)
-	    screen_putf(_("%s: Cannot close file - %s\n"), filename, errstr);
+	    warn(_("%s: Cannot close file - %s\n"), quotearg(filename), errstr);
 	return;
     }
     if (fclose(file) < 0)
-	screen_putf(_("%s: Cannot close file - %s\n"), filename, errstr);
+	warn(_("%s: Cannot close file - %s\n"), quotearg(filename), errstr);
 }
 
 bool
 set_share_dir(const char *dir)
 {
-    free(share_dir);
+    char *tmp = share_dir;
+    /* set first, because we may be called as set_share_dir(share_dir); */
     share_dir = (dir == NULL ? NULL : xstrdup(dir));
+    free(tmp);
     if (!filelist_create(share_dir))
     	return false;
     if (hub_state >= DC_HUB_LOGGED_IN && !send_my_info())
@@ -613,14 +620,9 @@ search_input_available(void)
 	return;
     }
 
-    if (search_recvq->buf[search_recvq->cur-1] != '|') {
-        search_recvq->buf[search_recvq->cur-1] = '\0';
-    	screen_putf(_("Unterminated $SR in UDP packet, discarding.\n"));
-	/* XXX: dump package? */
-    } else {
-	search_recvq->buf[search_recvq->cur-1] = '\0';
-	handle_search_result(search_recvq->buf, search_recvq->cur);
-    }
+    dump_command("<==", search_recvq->buf, search_recvq->cur);
+    search_recvq->buf[search_recvq->cur-1] = '\0'; /* not strictly necessary */
+    handle_search_result(search_recvq->buf, search_recvq->cur);
     byteq_clear(search_recvq);
 }
 
@@ -751,7 +753,7 @@ enable_active(uint16_t port)
 	return false;
     }
 
-    warn(_("Listening on %s.\n"), sockaddr_in_str(&addr));
+    screen_putf(_("Listening on %s.\n"), sockaddr_in_str(&addr));
     listen_port = ntohs(addr.sin_port);
     FD_SET(listen_socket, &read_fds);
     return true;
@@ -793,7 +795,7 @@ handle_listen_connection(void)
 	return;
     }
 
-    screen_putf(_("User from %s connected.\n"), sockaddr_in_str(&addr));
+    flag_putf(DC_DF_CONNECTIONS, _("User from %s connected.\n"), sockaddr_in_str(&addr));
     user_connection_new(&addr, socket);
 }
 
@@ -805,9 +807,11 @@ main (int argc, char **argv)
     char *tmpdir;
     struct sigaction sigact;
 
-#ifdef ENABLE_NLS
+    set_quoting_style(NULL, escape_quoting_style);
+    
     if (setlocale(LC_ALL, "") == NULL)
        warn(_("%s: Cannot set locale: %s\n"), argv[0], errstr);
+#ifdef ENABLE_NLS
     if (bindtextdomain(PACKAGE, LOCALEDIR) == NULL)
        warn(_("%s: Cannot bind message domain: %s\n"), argv[0], errstr);
     if (textdomain(PACKAGE) == NULL)
@@ -821,7 +825,7 @@ main (int argc, char **argv)
 
 	switch (c) {
 	case HELP_OPT:
-	    printf(_("Usage: %s [OPTION]...\n"), argv[0]);
+	    printf(_("Usage: %s [OPTION]...\n"), quotearg(argv[0]));
 	    puts(_("Start microdc, a command-line based Direct Connect client.\n"));
 	    printf(_("      --help     display this help and exit\n"));
 	    printf(_("      --version  output version information and exit\n"));
@@ -886,7 +890,7 @@ main (int argc, char **argv)
     my_speed = xstrdup("56Kbps");
     my_tag = xasprintf("%s V:%s", PACKAGE, VERSION);
     download_dir = xstrdup(".");
-    tmpdir = get_temp_dir();
+    tmpdir = tempdir();
     if (tmpdir == NULL) {
 	warn(_("Cannot find directory for temporary files - %s\n"), errstr);
 	goto cleanup;
@@ -970,7 +974,7 @@ cleanup:
     ptrv_foreach(our_searches, (PtrVForeachCallback) free_search_request);
     ptrv_free(our_searches);
 
-    hmap_foreach(user_conns, user_conn_cancel);
+    hmap_foreach_value(user_conns, user_conn_cancel);
     hmap_free(user_conns);
 
     if (our_filelist != NULL)
@@ -992,12 +996,12 @@ cleanup:
 
 	    if (stat(filename, &st) < 0) {
 		if (errno != ENOENT)
-	    	    screen_putf(_("%s: Cannot get file status - %s\n"), filename, errstr);
+	    	    warn(_("%s: Cannot get file status - %s\n"), quotearg(filename), errstr);
 		free(filename);
 		continue;
 	    }
     	    if (unlink(filename) < 0)
-		screen_putf(_("%s: Cannot remove file - %s\n"), filename, errstr);
+		warn(_("%s: Cannot remove file - %s\n"), quotearg(filename), errstr);
 	    free(filename);
 	}
 	ptrv_free(delete_files);
@@ -1010,12 +1014,12 @@ cleanup:
 
 	    if (stat(filename, &st) < 0) {
 		if (errno != ENOENT)
-	    	    screen_putf(_("%s: Cannot get file status - %s\n"), filename, errstr);
+	    	    warn(_("%s: Cannot get file status - %s\n"), quotearg(filename), errstr);
 		free(filename);
 		continue;
 	    }
     	    if (rmdir(filename) < 0)
-		screen_putf(_("%s: Cannot remove file - %s\n"), filename, errstr);
+		warn(_("%s: Cannot remove file - %s\n"), quotearg(filename), errstr);
 	    free(filename);
 	}
 	ptrv_free(delete_dirs);
