@@ -35,20 +35,24 @@
 #include "minmax.h"		/* Gnulib */
 #include "xstrndup.h"		/* Gnulib */
 #include "xvasprintf.h"		/* Gnulib */
+#include "quotearg.h"		/* Gnulib */
 #include "gettext.h"		/* Gnulib/GNU gettext */
 #define _(s) gettext(s)
 #define N_(s) gettext_noop(s)
 #include "microdc.h"
 #include "common/error.h"
 #include "common/strleftcmp.h"
-#include "common/common.h"
+#include "common/bksearch.h"
+#include "common/quoting.h"
 
 static void clear_rl();
 static void user_input(char *line);
 static int real_screen_vputf(const char *format, va_list args);
 static char *screen_prompt;
+static FILE *log_fh = NULL;
+char *log_filename = NULL;
 
-printf_fn_t screen_writer = real_screen_vputf;
+vprintf_fn_t screen_writer = real_screen_vputf;
 
 enum {
     SCREEN_NO_HISTORY,
@@ -58,177 +62,30 @@ enum {
     SCREEN_RL_CLEARED,		/* Readline has been cleared. Need to redisplay after. */
 } screen_state = SCREEN_NO_HISTORY;
 
-static char *
-quote_argument(const char *string, int rtype, int cs)
+bool
+set_log_file(const char *new_filename, bool verbose)
 {
-    const char *s;
-    char *result;
-    char *r;
-
-    if (cs == '"') {
-        result = xmalloc(2 * strlen(string) + 3);
-        r = result;
-        *r++ = '"';
-
-        for (s = string; *s; s++) {
-	    switch (*s) {
-	    case '"':
-	    case '\\':
-	        *r++ = '\\';
-	    default:
-	        *r++ = *s;
-	        break;
-            }
-        }
-
-        if (rtype != MULT_MATCH)
-            *r++ = '"';
-        *r = '\0';
-    } else {
-        result = xmalloc(2 * strlen(string) + 1);
-        r = result;
-
-        for (s = string; *s; s++) {
-	    switch (*s) {
-	    case ' ':
-	    case '\t':
-	    case '\n':
-	    case '"':
-	    case '\'':
-	    case '\\':
-	        *r++ = '\\';
-	        *r++ = *s;
-	        break;
-            default:
-	        *r++ = *s;
-	        break;
-            }
-        }
-
-        *r = '\0';
+    if (log_fh != NULL) {
+	if (fclose(log_fh) != 0)
+	    warn(_("%s: Cannot close file - %s\n"), quotearg(log_filename), errstr);
+	log_fh = NULL;
+	free(log_filename);
+	log_filename = NULL;
     }
-
-    return result;
-}
-
-static char *
-dequote_argument(const char *text, int quote_char, int count)
-{
-    const char *p;
-    char *ret;
-    char *r;
-    int quoted;
-
-    ret = xmalloc(strlen(text) + 1);
-    quoted = quote_char;
-    r = ret;
-    for (p = text; p-text < count && *p; p++) {
-	if (*p == '\\') { /* don't touch backslash and escaped chars */
-	    *r++ = *++p;
-	    if (p-text >= count && *p == '\0')
-		break;
-	    continue;
-	}
-	if (quoted != 0 && *p == quoted) {	/* close quote */
-	    quoted = 0;
-	    continue;
-	}
-	if (quoted == 0 && *p == '"') {	/* open quote */
-	    quoted = *p;
-	    continue;
-	}
-	*r++ = *p;
+    if (new_filename == NULL) {
+	if (verbose)
+	    screen_putf(_("No longer logging to file.\n"));
+	return true;
     }
-    *r = '\0';
-
-    return ret;
-}
-
-static const char *
-skip_word(const char *p)
-{
-    int quoted;
-
-    quoted = 0;
-    for (; *p; p++) {
-	if (*p == '\\') { /* don't touch backslash and escaped chars */
-	    p++;
-	    if (*p == '\0')
-		break;
-	    continue;
-	}
-	if (quoted != 0 && *p == quoted) {	/* Close quote. */
-	    quoted = 0;
-	    continue;
-	}
-	if (quoted == 0 && *p == '"') {	/* Open quote. */
-	    quoted = *p;
-	    continue;
-	}
-	if (quoted == 0 && (*p == ' ' || *p == '\t' || *p == '\n'))
-	    break;
+    log_fh = fopen(new_filename, "a");
+    if (log_fh == NULL) {
+	screen_putf(_("%s: Cannot open file for appending - %s\n"), quotearg(new_filename), errstr);
+	return false;
     }
-
-    return p;
-}
-
-char *
-get_argument(const char *p, int index, int count)
-{
-    char *space_chars = " \t\n";
-    const char *start;
-
-    if (p == NULL) /* necessary when no argument specified */
-	return NULL;
-    
-    /* Skip initial whitespace */
-    for (; *p && strchr(space_chars, *p) != NULL; p++);
-
-    /* Skip words prior */
-    for (; *p && index > 0; index--) {
-	p = skip_word(p);
-	for (; *p && strchr(space_chars, *p) != NULL; p++);
-    }
-    if (*p == '\0')
-	return NULL;
-
-    /* Skip 'count' number of words */
-    start = p;
-    for (; *p && count > 0; count--) {
-	p = skip_word(p);
-	if (count > 1)
-	    for (; *p && strchr(space_chars, *p) != NULL; p++);
-    }
-
-    return dequote_argument(start, 0, p-start);
-}
-
-int
-char_is_quoted(char *string, int index)
-{
-    bool escaped = false;
-    int c;
-
-    for (c = 0; c <= index; c++) {
-	if (escaped) {
-	    escaped = false;
-	    if (c >= index)
-		return 1;
-	} else if (string[c] == '"') {
-	    char quote = string[c];
-	    c++;
-	    for (; string[c] != '\0' && string[c] != quote; c++) {
-		if (string[c] == '\\' && string[c+1] != '\0')
-		    c++;
-	    }
-	    if (c >= index)
-		return 1;
-	} else if (string[c] == '\\') {
-	    escaped = true;
-	}
-    }
-
-    return 0;
+    log_filename = xstrdup(new_filename);
+    if (verbose)
+	screen_putf(_("Logging to `%s'.\n"), quotearg(new_filename));
+    return true;
 }
 
 /* Readline < 5.0 disables SA_RESTART on SIGWINCH for some reason.
@@ -257,19 +114,34 @@ attempted_completion(const char *text, int start, int end)
     char *dir_part = NULL;
     char *file_part = NULL;
     char *text_nq;
-    char quoted;
+    bool quoted;
 
     /* Determine if current word is quoted, and dequote. */
     quoted = char_is_quoted(rl_line_buffer, start)
-              && rl_line_buffer[start-1] == '"' ? '"' : '\0';
-    text_nq = dequote_argument(text, quoted, INT_MAX);
+              && rl_line_buffer[start-1] == '"';
+    text_nq = dequote_words(text, quoted, NULL);
    
     /* Do not attempt readline's default filename completion if our
      * completors fails to return any results.
      */
     rl_attempted_completion_over = 1;
 
-    completor = get_completor(rl_line_buffer, start, end, &type);
+    completor = default_completion_selector(rl_line_buffer, start, end, &type);
+    /* XXX: Better idea than DC_CPL_CHAIN: 
+     * Let the completor itself determine how to handle
+     * arguments. Is the dir/file_part separation really necessary?
+     * Cannot the completor itself do this itself? All that is
+     * necessary is rules before, and rules after. Shouldn't the
+     * completer return a complete set of results instead? And be
+     * passed a LList or something where results can be added?
+     * The first will always be the input result... But gcd (leading
+     * character) still need to be calculated.
+     */
+
+    if (type == DC_CPL_CHAIN) {
+	DCChainCompletionFunction chain_completor = completor;
+	completor = chain_completor(rl_line_buffer, start, end, &type);
+    }
     if (type == DC_CPL_FILE) {
 	file_part = strrchr(text_nq, '/');
 	if (file_part == NULL) {
@@ -285,7 +157,7 @@ attempted_completion(const char *text, int start, int end)
 	file_part = text_nq;
     }
 
-    if (completor != NULL) {
+    if (type == DC_CPL_NONE || completor != NULL) {
     	PtrV *completion_entries;
     	int state;
 	int lcd = INT_MAX;
@@ -321,7 +193,7 @@ attempted_completion(const char *text, int start, int end)
 		    if (c1 != c2)
 		    	break;
     	    	}
-	    	lcd = min(lcd, si);
+	    	lcd = MIN(lcd, si);
 	    }
 	}
 
@@ -356,7 +228,7 @@ attempted_completion(const char *text, int start, int end)
 		ce->str = tmp;
 	    }
 
-	    matches[0] = quote_argument(ce->str, SINGLE_MATCH, quoted);
+	    matches[0] = quote_word(ce->str, quoted, true);
 	    matches[1] = NULL;
 
 	    rl_completion_append_character = ce->input_char;
@@ -381,7 +253,7 @@ attempted_completion(const char *text, int start, int end)
 		tmp = xasprintf(ce->input_format, matches[0]);
 		free(matches[0]);
 	    }
-	    matches[0] = quote_argument(tmp, MULT_MATCH, quoted);
+	    matches[0] = quote_word(tmp, quoted, false);
 	    free(tmp);
 
 	    for (c = 0; c < state; c++) {
@@ -462,6 +334,8 @@ screen_putf(const char *format, ...)
 
     va_start(args, format);
     screen_writer(format, args);
+    if (log_fh != NULL)
+	vfprintf(log_fh, format, args);
     va_end(args);
 }
 
@@ -470,11 +344,12 @@ flag_putf(DCDisplayFlag flag, const char *format, ...)
 {
     va_list args;
 
-    if (display_flags & flag) {
-	va_start(args, format);
+    va_start(args, format);
+    if (display_flags & flag)
 	screen_writer(format, args);
-	va_end(args);
-    }
+    if (log_fh != NULL && log_flags & flag)
+	vfprintf(log_fh, format, args);
+    va_end(args);
 }
 
 /* This function is called by readline whenever the user has
@@ -486,6 +361,9 @@ user_input(char *line)
     /* Readline has already made way for us. */
     screen_state = SCREEN_NORMAL;
 
+    if (log_fh != NULL)
+	fprintf(log_fh, "> %s\n", line == NULL ? "(null)" : line);
+    
     if (line == NULL) {
         /* Ctrl+D was pressed on an empty line. */
         screen_putf("exit\n");
@@ -536,10 +414,11 @@ screen_finish(void)
     if (screen_state > SCREEN_NO_HISTORY) {
     	char *path;
 
+	/* Save history */
     	get_package_file("history", &path);
     	if (mkdirs_for_file(path, false) >= 0) {
 	    if (write_history(path) != 0)
-		warn(_("%s: Cannot write history - %s\n"), path, errstr);
+		warn(_("%s: Cannot write history - %s\n"), quotearg(path), errstr);
 	}
 	free(path);
 
@@ -552,6 +431,8 @@ screen_finish(void)
 
 	warn_writer = default_warn_writer;
         screen_state = SCREEN_NO_HISTORY;
+
+	set_log_file(NULL, false);
     }
 }
 
@@ -581,7 +462,7 @@ screen_prepare(void)
         using_history();
     	get_package_file("history", &path);
 	if (read_history(path) != 0 && errno != ENOENT)
-    	    warn(_("%s: Cannot read history - %s\n"), path, errstr);
+    	    warn(_("%s: Cannot read history - %s\n"), quotearg(path), errstr);
     }
     if (screen_state == SCREEN_NO_HANDLER) {
         rl_callback_handler_install(screen_prompt, user_input);
@@ -622,3 +503,29 @@ set_screen_prompt(const char *prompt, ...)
     va_end(args);
     rl_set_prompt(screen_prompt);
 }
+
+/* Look up completion alternatives from a sorted list using binary search.
+ */
+char *
+sorted_list_completion_generator(const char *base, int state, void *items,
+                                 size_t item_count, size_t item_size,
+                                 size_t key_offset)
+{
+    static const void *item;
+    static const void *last_item;
+
+    if (state == 0) {
+	if (!bksearchrange(base, items, item_count, item_size,
+	                   key_offset, (comparison_fn_t) strleftcmp,
+	                   &item, &last_item))
+	    return NULL;
+    }
+
+    if (item <= last_item) {
+	char *key = xstrdup(*(const char **) (((const char *) item) + key_offset));
+	item = (const void *) (((const char *) item) + item_size);
+	return key;
+    }
+
+    return NULL;
+}
