@@ -28,12 +28,12 @@
 #include <time.h>		/* ? */
 #include "xalloc.h"		/* Gnulib */
 #include "xstrndup.h"		/* Gnulib */
+#include "quotearg.h"		/* Gnulib */
 #include "gettext.h"		/* Gnulib/GNU gettext */
 #define _(s) gettext(s)
 #define N_(s) gettext_noop(s)
 #include "common/strbuf.h"
 #include "common/intutil.h"
-#include "common/common.h"
 #include "microdc.h"
 
 #define MAX_RESULTS_ACTIVE 10		/* Max number of search results to send to active users */
@@ -53,6 +53,22 @@ static char *extensions[] = { // NULL means match any extension
     /* CHECKSUM */ NULL,
 };
 
+static int
+compare_search_response(DCSearchResponse *sr1, DCSearchResponse *sr2)
+{
+    /* refcount is not derived from results - not compared. */
+    COMPARE_RETURN_FUNC(strcmp(sr1->userinfo->nick, sr2->userinfo->nick));
+    COMPARE_RETURN_FUNC(strcmp(sr1->filename, sr2->filename));
+    COMPARE_RETURN(sr1->filetype, sr2->filetype);
+    COMPARE_RETURN(sr1->filesize, sr2->filesize);
+    /* slots_free and slots_total may change - not compared. */
+    /* hub_name and hub_addr may be incomplete - not compared. */
+    /* XXX: when/if multiple hub connections are supported, we probably need
+     * to compare hub_name and/or hub_addr as well.
+     */
+    return 0;
+}
+
 static void
 search_string_new(DCSearchString *sp, const char *p, int len)
 {
@@ -303,7 +319,7 @@ perform_inbound_search(DCSearchSelection *data, DCUserInfo *ui, struct sockaddr_
 
     if (curresults > 0) {
         if (ui != NULL) {
-            flag_putf(DC_DF_DEBUG, _("Sent %d/%d search results to %s\n"), curresults, maxresults, ui->nick);
+            flag_putf(DC_DF_DEBUG, _("Sent %d/%d search results to %s\n"), curresults, maxresults, quotearg(ui->nick));
         } else {
             flag_putf(DC_DF_DEBUG, _("Sent %d/%d search results to %s\n"), curresults, maxresults, sockaddr_in_str(addr));
         }
@@ -340,6 +356,16 @@ parse_search_response(char *buf, uint32_t len)
     if (ui == NULL)
     	return NULL; /* Invalid $SR message: Unknown user. */
 
+    /* Why look for `/' here?
+     * A search result looks like this:
+     * Directories: $SR <nick><0x20><directory><0x20><free slots>/<total slots><0x05><Hubname><0x20>(<Hubip:port>)
+     * Files:       $SR <nick><0x20><filename><0x05><filesize><0x20><free slots>/<total slots><0x05><Hubname><0x20>(<Hubip:port>)
+     * In the directories case, the only way to determine where the file
+     * name ends is to look for `/' because it delimits the free slots and
+     * total slots value. It is assumed that file names cannot contain the
+     * slash (which is true for both *nix and windows).
+     */
+
     filename = strsep(&buf, "/");
     if (filename == NULL)
     	return NULL; /* Invalid $SR message: Missing filename. */
@@ -377,9 +403,12 @@ parse_search_response(char *buf, uint32_t len)
     *token = '\0';
     buf = token+1;
     token = strchr(buf, ')');
-    if (token == NULL)
-    	return NULL; /* Invalid $SR message: Missing hub address. */
-    *token = '\0';
+    /* NeoModus DirectConnect 2.20 doesn't appear to add `)|' to $SR
+     * messages. Or some other client which puts 'DC V:2.20' in the
+     * client tag.
+     */
+    if (token != NULL)
+        *token = '\0';
 
     token = strchr(buf, ':');
     if (token != NULL) {
@@ -491,7 +520,7 @@ add_search_request(char *args)
 	    break;
     }
 
-    if (now < 0 && time(&now) < 0) {
+    if (time(&now) == (time_t) -1) {
 	warn(_("Cannot get current time - %s\n"), errstr);
 	free(sel.patterns);
 	return false;
@@ -535,14 +564,16 @@ handle_search_result(char *buf, uint32_t len)
     uint32_t c;
     time_t now;
 
-    if (time(&now) < 0) {
+    if (time(&now) == (time_t) -1) {
 	warn(_("Cannot get current time - %s\n"), errstr);
 	return;
     }
 
     sr = parse_search_response(buf, len);
-    if (sr == NULL)
+    if (sr == NULL) {
+        warn(_("Unterminated or invalid $SR, discarding: %s\n"), quotearg_mem(buf, len));
     	return;
+    }
 
     for (c = 0; c < our_searches->cur; c++) {
     	DCSearchRequest *sd = our_searches->buf[c];
@@ -552,15 +583,42 @@ handle_search_result(char *buf, uint32_t len)
 
    	if (match_selection_against_response(&sd->selection, sr)) {
 	    char *fmt;
+	    uint32_t d;
 
-	    ptrv_append(sd->responses, sr);
-	    fmt = ngettext("Added search result to search %d (now %d result).\n",
-			   "Added search result to search %d (now %d results).\n",
-			   sd->responses->cur);
-	    flag_putf(DC_DF_SEARCH_RESULTS, fmt, c+1, sd->responses->cur);
-	    sr->refcount++;
+	    /* This is slow, but better than dupes... */
+	    for (d = 0; d < sd->responses->cur; d++) {
+		if (compare_search_response(sd->responses->buf[d], sr) == 0) {
+		    screen_putf(_("Result has been added earlier to search %d.\n"), c+1);
+		    break;
+		}
+	    }
+	    if (d >= sd->responses->cur) {
+		ptrv_append(sd->responses, sr);
+		fmt = ngettext("Added result to search %d (now %d result).\n",
+			       "Added result to search %d (now %d results).\n",
+			       sd->responses->cur);
+		flag_putf(DC_DF_SEARCH_RESULTS, fmt, c+1, sd->responses->cur);
+		sr->refcount++;
+	    }
 	}
     }
 
     free_search_response(sr);
 }
+
+char *
+search_selection_to_string(DCSearchSelection *sr)
+{
+    StrBuf *out;
+    uint32_t c;
+
+    out = strbuf_new_from_char('"');
+    for (c = 0; c < sr->patterncount; c++) {
+        if (c != 0)
+            strbuf_append_char(out, ' ');
+        strbuf_append_substring(out, sr->patterns[c].str, 0, sr->patterns[c].len);
+    }
+    strbuf_append_char(out, '"');
+
+    return strbuf_free_to_string(out);
+}
