@@ -36,12 +36,12 @@
 #include "xvasprintf.h"		/* Gnulib */
 #include "xalloc.h"		/* Gnulib */
 #include "minmax.h"		/* Gnulib */
+#include "quotearg.h"		/* Gnulib */
 #include "gettext.h"		/* Gnulib/GNU gettext */
 #define _(s) gettext(s)
 #define N_(s) gettext_noop(s)
 #include "common/error.h"
 #include "common/intutil.h"
-#include "common/common.h"
 #include "microdc.h"
 
 #define LOCK_STRING "ABCABCABCABCABCABCABCABCABCABCABCA"
@@ -112,7 +112,7 @@ handle_main_msg(IPC *ipc, uint32_t id, void *userdata)
     //	ipc_get(ipc, id, IPC_END);
     //	return ipc_put(ipc, id, IPC_INT64, transfer_pos, IPC_END);
     default:
-        screen_putf(_("Received message %d from main process.\n"), id);
+        warn(_("Received unknown message %d from main process, shutting down.\n"), id);
 	user_running = false;
         return false;
     }
@@ -131,7 +131,7 @@ handle_main_ipc_error(IPC *ipc, IPCError error, int result, void *userdata)
     	warn_socket_error(result, false, _("main process"));
 	break;
     case IPC_ERROR_DATA:
-    	screen_putf(_("Logical communication error with main process.\n"));
+    	warn(_("Logical communication error with main process.\n"));
     	break;
     }
     user_running = false;
@@ -205,7 +205,7 @@ download_next_file(void)
     if (!ipc_get(ipc, DC_MSG_NEXT_DOWNLOAD, IPC_STR, &dl_file, IPC_STR, &dl_base, IPC_INT32, &file_flag, IPC_END))
     	return false;
     if (dl_file == NULL) {
-        screen_putf(_("No more files to download.\n"));
+        flag_putf(DC_DF_CONNECTIONS, _("No more files to download.\n"));/*XXX:move main.c?*/
         user_running = false;
 	return false;
     }
@@ -264,7 +264,7 @@ nick_validate(const char *str)
     if (!ipc_get(ipc, DC_MSG_VALIDATE_NICK, IPC_BOOL, &reply, IPC_END))
     	return false;
     if (!reply) {
-	screen_putf(_("User %s not on hub, or too many connections to user.\n"), str);
+	warn(_("User %s not on hub, or too many connections to user.\n"), quotearg(str));/*XXX: move main.c */
 	user_running = false;
     	return false;
     }
@@ -293,24 +293,26 @@ open_download_file(void)
 
     transfer_file/*DL*/ = open/*64*/(localfile, O_CREAT|O_EXCL|O_WRONLY, 0644);
     if (transfer_file/*DL*/ < 0) {
-	warn(_("%s: Cannot open file for writing - %s\n"), localfile, errstr);
+	warn(_("%s: Cannot open file for writing - %s\n"), quotearg(localfile), errstr);
 	//free(localfile); localfile = NULL;
 	//free(file_name/*DL*/); file_name/*DL*/ = NULL;
-	/* FIXME: should really try to download file later instead... */
+	/* XXX: should really try to download file later instead... */
         //if (!ipc_put(ipc, DC_MSG_TRANSFER_FAILED/*DL*/, IPC_INT32, file_flag, IPC_END))
         //   return;
 	return download_next_file();
     }
 
     if (file_size == 0) {
-	warn(_("File %s is empty, not starting download.\n"), file_name/*DL*/);
+	warn(_("File %s is empty, not starting download.\n"), quotearg(file_name/*DL*/));
 	if (close(transfer_file/*DL*/) < 0)
-	    warn(_("%s: Cannot close file - %s\n"), file_name/*DL*/, errstr);
+	    warn(_("%s: Cannot close file - %s\n"), quotearg(file_name/*DL*/), errstr);
 	transfer_file/*DL*/ = -1;
     } else {
         if (!user_putf("$Send|"))
 	    return false; /* maybe free file_name here? */
-        warn(_("Starting download for file %s, %" PRIu64 " bytes\n"), file_name/*DL*/, file_size);
+        /* XXX: this could be DC_DF_DOWNLOAD as well - but that is only for completed downloads!? */
+        /* XXX: this could move to main:DC_MSG_TRANSFER_STARTING */
+        flag_putf(DC_DF_CONNECTIONS, _("Starting download for file %s, %" PRIu64 " bytes\n"), quotearg(file_name/*DL*/), file_size);
     }
 
     /* Let the main process think we are starting the download even if
@@ -363,7 +365,7 @@ open_upload_file(const char *str, uint64_t offset)
     	return false;
 
     if (uploadfile == NULL) {
-        screen_putf(_("Cannot resolve upload file `%s' - invalid filename or no sharedir\n"), newname);
+        warn(_("Cannot resolve upload file `%s' - invalid filename or no sharedir\n"), quotearg(newname));/*XXX:move main.c */
 	free(newname);
 	if (!user_putf("$Error File Not Available|"))
 	    return false;
@@ -377,14 +379,14 @@ open_upload_file(const char *str, uint64_t offset)
 
     /* Get file status and check offset. */
     if (stat(file_name/*UL*/, &st) < 0) {
-        screen_putf(_("%s: Cannot get file status - %s\n"), file_name/*UL*/, errstr);
+        warn(_("%s: Cannot get file status - %s\n"), quotearg(file_name/*UL*/), errstr);
 	//free(file_name/*UL*/); file_name/*UL*/ = NULL;
 	if (!user_putf("$Error File Not Available|"))
 	    return false;
         return true;
     }
     if (offset > st.st_size) {
-	screen_putf(_("%s: Offset %" PRId64 " out of range\n"), file_name/*UL*/, offset);
+	warn(_("%s: Offset %" PRId64 " out of range\n"), quotearg(file_name/*UL*/), offset);
 	//free(file_name/*UL*/); file_name/*UL*/ = NULL;
 	if (!user_putf("$Error Offset out of range|"))
 	    return false;
@@ -394,18 +396,18 @@ open_upload_file(const char *str, uint64_t offset)
     /* Open file and seek to desired position. */
     transfer_file/*UL*/ = open/*64*/(file_name/*UL*/, O_RDONLY);
     if (transfer_file/*UL*/ < 0) {
-	screen_putf(_("%s: Cannot open file for reading - %s\n"), file_name/*UL*/, errstr);
+	warn(_("%s: Cannot open file for reading - %s\n"), quotearg(file_name/*UL*/), errstr);
 	//free(file_name/*UL*/); file_name/*UL*/ = NULL;
 	if (!user_putf("$Error File Not Available|"))
 	    return false;
         return true;
     }
     if (offset != 0 && lseek/*64*/(transfer_file/*UL*/, offset, SEEK_SET) < 0) {
-    	screen_putf(_("%s: Cannot seek in file - %s\n"), file_name/*UL*/, errstr);
+    	warn(_("%s: Cannot seek in file - %s\n"), quotearg(file_name/*UL*/), errstr);
 	if (!user_putf("$Error File Not Available|"))
 	    return false;
 	if (close(transfer_file/*UL*/) < 0)
-	    warn(_("%s: Cannot close file - %s\n"), file_name/*UL*/, errstr);
+	    warn(_("%s: Cannot close file - %s\n"), quotearg(file_name/*UL*/), errstr);
 	transfer_file/*UL*/ = -1;
 	//free(file_name/*UL*/); file_name/*UL*/ = NULL;
 	return true;
@@ -460,7 +462,7 @@ user_handle_command(char *buf, uint32_t len)
 	if (res < len) {
 	    warn_file_error(res, true, file_name/*DL*/);
 	    if (close(transfer_file/*DL*/) < 0)
-	    	warn(_("%s: Cannot close file - %s\n"), file_name/*DL*/, errstr);
+	    	warn(_("%s: Cannot close file - %s\n"), quotearg(file_name/*DL*/), errstr);
 	    transfer_file/*DL*/ = -1;
 	    //free(localfile); localfile = NULL;
 	    //free(file_name/*DL*/); file_name/*DL*/ = NULL;
@@ -473,14 +475,14 @@ user_handle_command(char *buf, uint32_t len)
 	data_size/*DL*/ -= len;
     	send_user_status(file_pos);
     	if (file_pos == file_size) {
-	    flag_putf(DC_DF_DOWNLOAD, _("%s: Download complete.\n"), base_name(file_name/*DL*/));
+	    flag_putf(DC_DF_DOWNLOAD, _("%s: Download complete.\n"), quotearg(base_name(file_name/*DL*/)));/* XXX: maybe move main.c DC_MSG_TRANSFER_FINISHED?*/
 	    if (close(transfer_file/*DL*/) < 0)
-	    	warn(_("%s: Cannot close file - %s\n"), file_name/*DL*/, errstr);
+	    	warn(_("%s: Cannot close file - %s\n"), quotearg(file_name/*DL*/), errstr);
             if (file_flag == DC_TF_NORMAL) {
                 char *tmp;
                 tmp = xstrndup(localfile, strlen(localfile)-5);
                 if (rename(localfile, tmp) < 0)
-                    warn(_("%s: Cannot rename file to `%s' - %s\n"), localfile, tmp, errstr);
+                    warn(_("%s: Cannot rename file to `%s' - %s\n"), quotearg_n(0, localfile), quotearg_n(1, tmp), errstr);
                 free(tmp);
             }
             //free(localfile); localfile = NULL;
@@ -514,7 +516,7 @@ user_handle_command(char *buf, uint32_t len)
 	    return;
         key = memmem(buf+6, len-6, " Pk=", 4);
         if (key == NULL) {
-            screen_putf(_("Invalid $Lock message: Missing Pk value\n"));
+            warn(_("Invalid $Lock message: Missing Pk value\n"));
 	    key = buf+len;
 	}
         key = decode_lock(buf+6, key-buf-6, DC_CLIENT_BASE_KEY);
@@ -558,7 +560,7 @@ user_handle_command(char *buf, uint32_t len)
 
 	token = strtok(NULL, " ");
 	if (token == NULL) {
-	    screen_putf(_("Invalid $Direction message: Missing challenge parameter\n"));
+	    warn(_("Invalid $Direction message: Missing challenge parameter\n"));
 	    user_running = false;
 	    return;
 	}
@@ -605,7 +607,7 @@ user_handle_command(char *buf, uint32_t len)
 	    return;
     	key = decode_lock(LOCK_STRING, LOCK_STRING_LEN, DC_CLIENT_BASE_KEY);
 	if (strcmp(buf+5, key) != 0)
-	    screen_putf(_("Invalid $Key message: Incorrect key, ignoring\n"));
+	    warn(_("Invalid $Key message: Incorrect key, ignoring\n"));
 	free(key);
     	if (our_dir == DC_DIR_SEND) {
 	    user_state = DC_USER_GET;
@@ -624,12 +626,12 @@ user_handle_command(char *buf, uint32_t len)
 	}
 	token = strtok(buf+5, "$");
 	if (token == NULL) {
-	    screen_putf(_("Invalid $Get message: Missing offset, assuming start\n"));
+	    warn(_("Invalid $Get message: Missing offset, assuming start\n"));
 	    offset = 0;
 	} else {
 	    token = strtok(NULL, ""); /* Cannot fail! */
 	    if (!parse_uint64(token, &offset)) {
-	    	screen_putf(_("Invalid $Get message: Offset not integer\n"));
+	    	warn(_("Invalid $Get message: Offset not integer\n"));
 		user_running = false;
 		return;
 	    }
@@ -644,7 +646,7 @@ user_handle_command(char *buf, uint32_t len)
 	    if (!ipc_put(ipc, DC_MSG_TRANSFER_FINISHED/*UL*/, IPC_INT32, 0, IPC_END))
 	    	return;
 	    if (close(transfer_file/*UL*/) < 0)
-		warn(_("%s: Cannot close file - %s\n"), file_name/*UL*/, errstr);
+		warn(_("%s: Cannot close file - %s\n"), quotearg(file_name/*UL*/), errstr);
 	    transfer_file/*UL*/ = -1;
 	}
     	open_upload_file(buf+5, offset); /* Changes state */
@@ -652,7 +654,7 @@ user_handle_command(char *buf, uint32_t len)
     else if (len == 9 && strcmp(buf, "$MaxedOut") == 0) {
     	if (!check_state(buf, DC_USER_FILE_LENGTH))
 	    return;
-	screen_putf(_("User is maxed out.\n"));
+	warn(_("User is maxed out.\n")); /*XXX: or DC_DF_DOWNLOAD?*/
 	user_running = false;
     }
     else if (len >= 12 && strncmp(buf, "$FileLength ", 12) == 0) {
@@ -668,16 +670,16 @@ user_handle_command(char *buf, uint32_t len)
     else if (len >= 7 && strncmp(buf, "$Error ", 7) == 0) {
 	if (user_state == DC_USER_FILE_LENGTH) {
     	    if (strcmp(buf+7, "File Not Available") == 0) {
-	    	screen_putf("%s: File not available on remote\n", file_name/*DL*/);
+	    	warn("%s: File not available on remote\n", quotearg(file_name/*DL*/));
 		if (close(transfer_file/*DL*/) < 0)
-	    	    warn(_("%s: Cannot close file - %s\n"), file_name/*DL*/, errstr);
+	    	    warn(_("%s: Cannot close file - %s\n"), quotearg(file_name/*DL*/), errstr);
 		// free(file_name/*DL*/); file_name/*UL*/ = NULL;
 		transfer_file/*DL*/ = -1;
 		download_next_file();
 	    }
 	    return;
 	}
-	screen_putf(_("Received error from user: %s\n"), buf+7);
+	warn(_("Received error from user: %s\n"), quotearg(buf+7));
 	user_running = false;
     }
     else if (len >= 5 && strncmp(buf, "$Send ", 5) == 0) {
@@ -765,7 +767,7 @@ user_now_writable(void)
 	user_state = DC_USER_MYNICK;
 
     	if (we_connected) {
-	    screen_putf(_("Connected to user.\n"));
+	    flag_putf(DC_DF_CONNECTIONS, _("Connected to user.\n"));/*XXX: move where!?*/
 	    if (!user_putf("$MyNick %s|", my_nick))
 		return;
     	    if (!user_putf("$Lock %s Pk=%s|", LOCK_STRING, LOCK_PK_STRING))
@@ -783,7 +785,7 @@ user_now_writable(void)
 	    if (res < block) {
 	    	warn_file_error(res, false, file_name/*UL*/);
 		if (close(transfer_file/*UL*/) < 0)
-		    warn(_("%s: Cannot close file - %s\n"), file_name/*UL*/, errstr);
+		    warn(_("%s: Cannot close file - %s\n"), quotearg(file_name/*UL*/), errstr);
 		transfer_file/*UL*/ = -1;
 		//free(file_name/*UL*/); file_name/*UL*/ = NULL;
 		user_running = false;
@@ -803,10 +805,10 @@ user_now_writable(void)
   	send_user_status(file_pos - user_sendq->cur);
 
 	if (file_pos == file_size && user_sendq->cur == 0) {
-	    flag_putf(DC_DF_UPLOAD, _("%s: Upload complete.\n"), base_name(file_name/*UL*/));
+	    flag_putf(DC_DF_UPLOAD, _("%s: Upload complete.\n"), base_name(file_name/*UL*/));/*XXX: maybe move main.c DC_MSG_TRANSFER_FINISHED? */
     	    FD_CLR(user_socket, &user_write_fds);
 	    if (close(transfer_file/*UL*/) < 0)
-	    	warn(_("%s: Cannot close file - %s\n"), file_name/*UL*/, errstr);
+	    	warn(_("%s: Cannot close file - %s\n"), quotearg(file_name/*UL*/), errstr);
 	    transfer_file = -1;
     	    free(file_name/*UL*/); file_name/*UL*/ = NULL;
 	    if (!ipc_put(ipc, DC_MSG_TRANSFER_FINISHED/*UL*/, IPC_INT32, 0, IPC_END))
@@ -999,26 +1001,26 @@ user_connection_new(struct sockaddr_in *addr, int user_socket)
     pid_t pid;
 
     if (socketpair(AF_LOCAL, SOCK_STREAM, 0, main_sockets) < 0) {
-    	screen_putf(_("Cannot create socket pair - %s\n"), errstr);
+    	warn(_("Cannot create socket pair - %s\n"), errstr);
 	return NULL;
     }
 
     pid = fork();
     if (pid < 0) {
-    	screen_putf(_("Cannot create process - %s\n"), errstr);
+    	warn(_("Cannot create process - %s\n"), errstr);
 	if (close(main_sockets[0]) < 0)
-	    screen_putf(_("Cannot close socket - %s\n"), errstr);
+	    warn(_("Cannot close socket - %s\n"), errstr);
 	if (close(main_sockets[1]) < 0)
-	    screen_putf(_("Cannot close socket - %s\n"), errstr);
+	    warn(_("Cannot close socket - %s\n"), errstr);
 	return NULL;
     }
     if (pid == 0)
 	user_main(main_sockets, addr, user_socket);
 
     if (close(main_sockets[1]) < 0)
-    	screen_putf(_("Cannot close socket - %s\n"), errstr);
+    	warn(_("Cannot close socket - %s\n"), errstr);
     if (user_socket >= 0 && close(user_socket) < 0)
-    	screen_putf(_("Cannot close socket - %s\n"), errstr);
+    	warn(_("Cannot close socket - %s\n"), errstr);
 
     uc = xmalloc(sizeof(DCUserConn));
     uc->pid = pid;
@@ -1039,7 +1041,7 @@ user_disconnect(DCUserConn *uc)
 {
     int status;
 
-    screen_putf(_("User %s: Shutting down user connection.\n"), uc->name);
+    flag_putf(DC_DF_CONNECTIONS, _("User %s: Shutting down user connection.\n"), quotearg(uc->name)); /* XXX: move where? */
     ipc_free(uc->ipc);
     if (uc->info != NULL) {
     	int c;
