@@ -24,10 +24,13 @@
 #include <unistd.h>		/* POSIX */
 #include <fcntl.h>		/* ? */
 #include "xalloc.h"		/* Gnulib */
+#include "xstrndup.h"		/* Gnulib */
 #include "xvasprintf.h"		/* Gnulib */
+#include "quotearg.h"		/* Gnulib */
 #include "gettext.h"		/* Gnulib/GNU gettext */
 #define _(s) gettext(s)
 #define N_(s) gettext_noop(s)
+#include "common/comparison.h"
 #include "microdc.h"
 
 /* Move to lib !? */
@@ -40,6 +43,7 @@ ilog10(uint64_t c)
     return r;
 }
 
+/* XXX: try to get rid of screen_putf from this function. */
 int
 mkdirs_for_file(char *filename, bool deletedirs)
 {
@@ -52,11 +56,11 @@ mkdirs_for_file(char *filename, bool deletedirs)
         *t = '\0';
 	if (stat(filename, &st) < 0) {
 	    if (errno != ENOENT) {
-	    	screen_putf(_("%s: Cannot get file status - %s\n"), filename, errstr);
+	    	screen_putf(_("%s: Cannot get file status - %s\n"), quotearg(filename), errstr);
 		return -1;
 	    } else {
         	if (mkdir(filename, 0777) < 0) {
-    	    	    screen_putf(_("%s: Cannot create directory - %s\n"), filename, errstr); 
+    	    	    screen_putf(_("%s: Cannot create directory - %s\n"), quotearg(filename), errstr); 
 		    return -1;
 		} else if (deletedirs) {
 	    	    if (ptrv_find(delete_dirs, filename, (comparison_fn_t) strcmp) < 0)
@@ -134,35 +138,40 @@ sockaddr_in_str(struct sockaddr_in *addr)
     return buffer;
 }
 
-bool
-direxists(const char *dir)
+PtrV *
+wordwrap(char *str, size_t len, size_t first_width, size_t other_width)
 {
-    struct stat sb;
-    return stat(dir, &sb) == 0 && S_ISDIR(sb.st_mode);
-}
+    size_t width;
+    PtrV *out;
 
-/* This does in no way replace the mkdtemp function provided by Gnulib.
- * mkdtemp still needs to know which directory to create the temporary
- * directory inside - it doesn't figure that out by itself.
- */
-char *
-get_temp_dir(void)
-{
-    char *tmpdir;
+    out = ptrv_new();
+    width = first_width;
+    while (len > width) {
+        size_t c, d;
 
-    if (getuid() == geteuid() && getgid() == getegid()) {
-	tmpdir = getenv("TMPDIR");
-	if (tmpdir != NULL && direxists(tmpdir))
-	    return tmpdir;
+        if (str[width] == ' ') {
+            ptrv_append(out, xstrndup(str, width));
+            for (c = width+1; c < len && str[c] == ' '; c++);
+            str += c;
+            len -= c;
+        } else {
+
+            for (c = width; c > 0 && str[c-1] != ' '; c--);
+            for (d = width+1; d < len && str[d] != ' '; d++);
+            if (d-c <= width) {
+                ptrv_append(out, xstrndup(str, c-1));
+                str += c;
+                len -= c;
+            } else {
+                ptrv_append(out, xstrndup(str, width));
+                str += width;
+                len -= width;
+            }
+        }
+        width = other_width;
     }
-    tmpdir = P_tmpdir;
-    if (tmpdir != NULL && direxists(tmpdir))
-	return tmpdir;
+    if (len > 0)
+        ptrv_append(out, xstrndup(str, len));
 
-    tmpdir = "/tmp";
-    if (direxists(tmpdir))
-	return tmpdir;
-
-    errno = ENOENT;
-    return NULL;
+    return out;
 }
